/** 
 * \file libdbg.h
 * \brief Header file with debug macroses and definition of debug functions.
 */

#ifndef DEBUG__H
#define DEBUG__H

#include <stdio.h>
#include <syslog.h>

#define XSTRINGIZE(s)           #s
#define STRINGIZE(s)            XSTRINGIZE(s)

/** \brief Structure with debug descriptor.
 *
 */
typedef struct
{
  char prefix[33]; /*!< Prefix for logging. */
  FILE *file;      /*!< Log file. */
  int   level;     /*!< Debug level. */
} log_desc_t;


#if !defined (NO_DEBUG) && defined (DBG_PREFIX)
    #define PREFIX_TO_VARIABLE2(prefix) prefix##_log_dsc
    #define PREFIX_TO_VARIABLE(prefix) PREFIX_TO_VARIABLE2(prefix)
    #define DECLARE_DBG_DSC(prefix) log_desc_t* PREFIX_TO_VARIABLE2(prefix)
    
    #define PREFIX_TO_CONSTR_FUNC2(prefix) static void prefix##_log_init(void)
    #define PREFIX_TO_CONSTR_FUNC(prefix) PREFIX_TO_CONSTR_FUNC2(prefix)

    #define PREFIX_TO_DESTR_FUNC2(prefix) static void prefix##_log_deinit(void)
    #define PREFIX_TO_DESTR_FUNC(prefix) PREFIX_TO_DESTR_FUNC2(prefix)

    DECLARE_DBG_DSC(DBG_PREFIX);
    
    #define ERR(format, ...) log_print(PREFIX_TO_VARIABLE(DBG_PREFIX), LOG_ERR, "ERR: "format"\n", ##__VA_ARGS__)

    #define WARN(format, ...) log_print(PREFIX_TO_VARIABLE(DBG_PREFIX), LOG_WARNING, "WARN: "format"\n", ##__VA_ARGS__)

    #define MSG(format, ...) log_print(PREFIX_TO_VARIABLE(DBG_PREFIX), LOG_INFO, format"\n", ##__VA_ARGS__)

    #define DBGL(level, format, ...) log_print(PREFIX_TO_VARIABLE(DBG_PREFIX), level, "DBG: %s: "format"\n", __FUNCTION__, ##__VA_ARGS__)

    #define DBG(format, ...) DBGL(LOG_DEBUG, format, ##__VA_ARGS__)
    
    #define DBG_START log_dbg_conf( &PREFIX_TO_VARIABLE(DBG_PREFIX), STRINGIZE(DBG_PREFIX) )

    #define DBG_STOP log_free( PREFIX_TO_VARIABLE(DBG_PREFIX) )
    
    #define DBG_LIB_CONSTR __attribute__((constructor)) PREFIX_TO_CONSTR_FUNC(DBG_PREFIX){\
                             DBG_START;}

    #define DBG_LIB_DESTR __attribute__((destructor)) PREFIX_TO_DESTR_FUNC(DBG_PREFIX){\
                             DBG_STOP;}
#else
  /** \brief Print error message */
  #define ERR(format, ...) fprintf( stderr, "Error: "format"\n", ##__VA_ARGS__)

  /** \brief Print warning message */
  #define WARN(format, ...) fprintf( stderr, "Warning: "format"\n", ##__VA_ARGS__)

  /** \brief Print log message */
  #define MSG(format, ...) fprintf( stdout, format"\n", ##__VA_ARGS__)

  /** \brief Print debug message with specific level */
  #define DBGL(level, format, ...)

  /** \brief Print debug message with default level */
  #define DBG(format, ...)

  /** \brief Configure libdbg and start/enable debugging */
  #define DBG_START 

  /** \brief Stop/disable debugging */
  #define DBG_STOP 

  /** \brief Debug messagging constructor to use in dynamic library */
  #define DBG_LIB_CONSTR
#endif

/** \brief Generic debugging macro */
#define LOG(format, ...) MSG(format, ##__VA_ARGS__)

/** \brief Function to initialize log descriptor structure.
 *  \param d Pointer to pointer to log descriptor.
 *  \param level Logging level.
 *  \param name Pointer to string with name of log file, sys means syslog, 
                NULL means stdout. 
 *  \param prefix Pointer to string that contains prefix for logging. 
 *  \return 0 if there are no errors, 1 otherwise
 */
int log_init(log_desc_t **d, int level, const char *name, const char *prefix);

/** \brief Function to close log file and free memory with descriptor struct.
 *  \param d Pointer to log descriptor.
 */
void log_free     ( log_desc_t *d );

/** \brief Function to print format string to log.
 *  \param d Pointer to log descriptor.
 *  \param mlevel String logging level.
 *  \param format Format string.
 */
void log_print    ( log_desc_t *d, int mlevel, const char *format, ... );

/** \brief Function to configure log descriptor from env variable.
 *  \param d Pointer to pointer to log descriptor.
 *  \param prefix Pointer to string that contains prefix for logging. 
 */
void log_dbg_conf ( log_desc_t **d, const char* prefix ); 

#endif //DEBUG__H
