#ifndef LIBGETOPTW_COMMON__H
#define LIBGETOPTW_COMMON__H

#include "config.h"

/** \brief Set dry-run mode.
 *
 *  \param p Pointer to prog_params_t structure
 *  \param val Pointer to string that contains value to write. Unused. 
 *
 *  \note Both params are used only for format compatibility. Actually they are
 *        ignored
 *
 *  \return Always 0
 */
int dry_run_common(prog_params_t *p, const char *val);

/** \brief Print program version
 *  \param p Pointer to prog_params_t structure
 *  \param val Pointer to string that contains value to write. Unused. 
 *  \return Always 0
 */
int cli_version_common(prog_params_t *p, const char *val);

/** \brief Set verbose mode
 *  \param p Pointer to prog_params_t structure
 *  \param val Pointer to string that contains value to write. Unused. 
 *  \return Always 0
 */
int cli_verbose_common(prog_params_t *p, const char *var);

/**
 * \brief Indentation (tab) for use with multi-line option help.
 *
 * Define milti-line option strings as:
 *
 *     "Option help text\n" OPTTAB "continued."
 *
 * \note The current indent length is 35 spaces.
 */
#define OPTTAB "                                   "

/** \brief Print help message. This message describes all options of a programm
 *
 *  \param params Pointer to null-terminated param_t array
 *  \param usage The usage text
 */
void common_print_help( param_t *params, const char *usage );

/** \brief Print help message
 *  \return Always 0
 */
int cli_help_common(prog_params_t *p, const char *var);

/** \brief Turn the debug mode on optionally setting its level
 *
 *  When val is null (no option argument is given) the procedure
 *  increments the value of p->common.debug by 1. Otherwise the
 *  string value is parsed and the resulting number is then assigned.
 *
 *  \param p Pointer to prog_params_t structure
 *  \param val Pointer to the null-terminated string representing
               the debug level number. Optional
 *
 *  \return Zero (0) on success, non-zero on fail
 */
int debug_level_common( prog_params_t *p, const char *val );

/**
 * \defgroup libgetoptw_common_options
 * \brief Options common for all utilities
 * @{
 */

/** \brief Macro implements `--help` option
 *
 */ 
#define HELP_OPT_COMMON \
  { \
    .opt  = 'h', \
    .name = "help", \
    .help = "print help and exit", \
    .type = PARAM_CLI_ONLY, \
    .has_arg = no_argument, \
    .handler = cli_help_common, \
  }

/** \brief Macro implements `--version` option
 *
 */ 
#define VERSION_OPT_COMMON \
  { \
    .opt  = 'V', \
    .name = "version", \
    .help = "print version and exit", \
    .type = PARAM_CLI_ONLY, \
    .has_arg = no_argument, \
    .handler = cli_version_common, \
  }

/** \brief Macro implements `--verbose` option
 *
 */ 
#define VERBOSE_OPT_COMMON \
  { \
    .opt  = 'v', \
    .name = "verbose", \
    .help = "verbose mode", \
    .type = PARAM_CLI_ONLY, \
    .has_arg = no_argument, \
    .handler = cli_verbose_common, \
  }

/** \brief Macro implements `--dry-run` option
 *
 */ 
#define DRY_RUN_OPT_COMMON \
  { \
    .opt  = 'n', \
    .name = "dry-run", \
    .help = "check settings and exit", \
    .type = PARAM_CLI_ONLY, \
    .handler = dry_run_common, \
    .has_arg = no_argument, \
  }

/** \brief Macro implements `--debug` option
 *
 */ 
#define DEBUG_LEVEL_OPT_COMMON \
  { \
    .opt  = 'd', \
    .name = "debug", \
    .help = "turn the debug mode on optionally setting its level", \
    .type = PARAM_GENERAL, \
    .handler = debug_level_common, \
    .has_arg = optional_argument, \
  }

/** \brief Macro implements special empty option that *must* be in the end of
 *         any @ref param_t array
 *
 */ 
#define TERMINATING_OPT \
  { \
    .opt  = 0,    \
    .name = NULL, \
    .help = NULL, \
    .type = 0,    \
    .handler = NULL, \
    .has_arg = 0, \
  }
/**@}*/

#endif //LIBGETOPTW_COMMON__H
