#ifndef __CONFIG_H__
#define __CONFIG_H__

#include <getopt.h>

enum {
  PARAM_CLI_ONLY,     //!< DEPRECATED, only command line, not in config
  PARAM_GENERAL,      //!< general, just params
  PARAM_STREAM,       //!< DEPRECATED, per-stream options
};

/**
 * \brief Parameters common for all programs
 *
 * These struct filled automatically.
 */
typedef struct {
  int debug;         //!< debug level
  int dry_run;       //!< dry_run option value
  int verbose;       //!< verbose option value
  char *version;     //!< version option value
  char *progname;    //!< Program name
  char *productname; //!< Product name
} common_params_t;

struct prog_param;

/**
 * \brief Option descriptor
 *
 * User have to initialize this struct for each of his options.
 *
 * For common options there are predefined initalizers HELP_OPT_COMMON,
 * VERSION_OPT_COMMON, VERSION_OPT_COMMON and DRY_RUN_OPT_COMMON,
 * DEBUG_LEVEL_OPT_COMMON.
 */
typedef struct {
  char opt;    //!< Short option name
  const char *name;  //!< Long option name
  int has_arg; //!< has_arg as in getopt_long
  const char *arg_name;  //!< Long option name
  char *help;  //!< help message to printed when `-h` or `--help` supplied
  int type;    //!< Type of option. Deprecated, use PARAM_GENERAL.
  int (*handler)( struct prog_param *prog_params, const char *val ); //!< Option parsing callback
} param_t;

/**
 * \brief Common parameters + user data
 *
 * You pass this struct to common_proc_all_config
 */
typedef struct prog_param {
  common_params_t common;       //!< Common parameters, see common_params_t.
  param_t *opt_params;          //!< Pointer to NULL-terminated array 
                                //!< of command line options.
  void *data;                   //!< User define program data.
} prog_params_t;

/** 
 * \brief Calculate and return number of parameters 
 *        in NULL-terminated param_t array.
 *
 *  \param params Pointer to param_t array
 *  \return number of parameters
 */
unsigned int calc_nparam( param_t *params );

/** 
 * \brief Look for param_t variable by name of command-line option
 *
 * \param params Pointer to param_t array
 * \param name String that contains name of command-line option
 * \return pointer to param_t variable if it is exists, NULL otherwise 
 */
param_t *common_find_param( param_t *params, const char *name );

/** 
 * \brief Heavy function that parses command-line options.
 *
 * It traverse params array invoking callback for each of the option. 
 * Common options are filled automatically by this function.
 *
 * Parsed option values are stored in `prog_params` where `prog_params->common` are
 * used for common option and user defined options can be stored in struct saved
 * in `prog_params->data`.
 *
 *  \param prog_params Pointer to prog_params_t structure
 *  \param params Pointer to param_t array
 *  \param argc Count of command-line arguments
 *  \param argv command-line arguments
 *  \param err Pointer to error value
 *  \return 0 if there are no errors, error code otherwise. Use "err" for
 *            additional error description
 */
int common_proc_all_config( prog_params_t *prog_params, param_t *params, 
                            int argc, char **argv, int *err );
#endif
