#!/bin/sh

if [ -z "${__included_boot_funcs-}" ]; then
__included_boot_funcs=1

BOOTCONFDIR="${BOOTCONFDIR:-/etc/rootfs-utils}"
BOOTCONF="${BOOTCONF:-$BOOTCONFDIR/boot.conf}"

get_current_boot() {
    local rootdev="$(get_current_root_device ||:)"

    if [ -z "$rootdev" ]; then
        echo 'UNKNOWN'
        return 1
    fi

    if echo "$rootdev" | grep -q "$RAM_ROOT_MATCH"; then
	echo "RAM"
	return 0	
    fi
    
    if echo "$rootdev" | grep -q "$NAND_ROOT_MATCH"; then
        echo 'NAND'
        return 0
    else
        if get_mmc_device_name "$rootdev"; then
            return 0
        fi
    fi

    echo 'UNKNOWN'
    return 2
}

get_next_boot() {
    if fw_printenv "$BOOT_VAR" 2> /dev/null | grep -q "$SD_BOOT_MATCH";  then
        local mmcpart="$(get_uboot_mmc_part)"
        if [ -n "$mmcpart" ]; then
            get_device_name "$SDDEV$mmcpart"
            return 0
        else
            echo "ERROR: Failed to get the MMC boot part number. U-Boot env seems to be broken or the configuration is not correct..." >&2
        fi
    elif fw_printenv "$BOOT_VAR" 2> /dev/null | grep -q "$NAND_BOOT_MATCH"; then
        echo 'NAND'
        return 0
    fi

    echo 'UNKNOWN'
    return 2
}

set_next_boot_nand()
{
    local bootcmd="$(get_current_bootcmd_base)"
    bootcmd="$bootcmd$NAND_BOOT_CMD"

    __fw_setenv "$BOOT_VAR" "$bootcmd"
}

set_next_boot_mmc()
{
    local sdnum="${1:-1}"
    local bootcmd="$(get_current_bootcmd_base)"
    local mmcdev="$(get_root_device "SD$sdnum")"
    
    local mmcpart="$(get_nth_mmcpartnum $sdnum)"
    if [ -z "$mmcpart" ]; then
        echo "ERROR: MMC part is not defined for SD$sdnum" >&2
        return 1
    fi
    
    mmccmd="$(printf "$MMC_BOOT_CMD" $mmcpart)"
    bootcmd="$bootcmd$mmccmd"

    __fw_setenv "$BOOT_VAR" "$bootcmd"

    if [ -n "$MMCPART_VAR" ]; then
        local mmcpartnum="$(printf "$MMCPART_SET" "$mmcpart")"
        __fw_setenv "$MMCPART_VAR" "$mmcpartnum"
    fi

    if [ -n "$MMCROOT_VAR" ]; then
        __fw_setenv "$MMCROOT_VAR" "$mmcdev"
    fi
}


if [ -n "$BOOTCONF" -a -e "$BOOTCONF" ]; then
    . "$BOOTCONF"
fi


BOOTNAMES="${BOOTNAMES-NAND SD1 SD2}"
NANDDEV="${NANDDEV:-/dev/mtdblock7}"
SDDEV="${SDDEV:-/dev/mmcblk0p}"
SDPARTLIST="${SDPARTLIST:-2 3}"
SDOVRLIST="${SDOVRLIST:-}"

BOOT_VAR="${BOOT_VAR:-bootcmd}"
RAM_ROOT_MATCH="${RAM_ROOT_MATCH:-ram0}"
NAND_ROOT_MATCH="${NAND_ROOT_MATCH:-ubi}"
NAND_BOOT_MATCH="${NAND_BOOT_MATCH:-run[[:space:]]\\+nandboot}"
NAND_BOOT_CMD="${NAND_BOOT_CMD:-run nandboot}"
SD_ROOT_MATCH="${SD_ROOT_MATCH:-$SDDEV\\([0-9]\\+\\)}"
SD_BOOT_MATCH="${SD_BOOT_MATCH:-run[[:space:]]\\+mmcboot}"
MMCPART_VAR="${MMCPART_VAR-mmcpart}"
MMCPART_MATCH="${MMCPART_MATCH:-\\([0-9]\\+\\)}"
MMCPART_SET="${MMCPART_SET:-%i}"
MMCROOT_VAR="${MMCROOT_VAR-root}"
MMC_BOOT_CMD="${MMC_BOOT_CMD:-run mmcboot}"

SDROOT_LABEL_PFX="${SDROOT_LABEL_PFX:-root-}"
SDHOME_LABEL="${SDHOME_LABEL:-home}"
SDCONF_LABEL="${SDCONF_LABEL:-conf}"
SDSTORAGE_LABEL="${SDSTORAGE_LABEL:-storage}"


is_bootname() {
    echo "$BOOTNAMES" | grep -qw "$1"
}

get_nth_mmcpartnum()
{
    local sdnum="$1"
    echo $SDPARTLIST | gawk -e "{ print \$$sdnum; }"
}

get_root_device()
{
    local name="$1"

    if ! is_bootname "$name"; then
        echo "Unknown boot device: $1" >&2
        return 1
    fi

    case "$name" in
        NAND)
            echo "$NANDDEV"
            return 0
            ;;
        SD*)
            if [ -f "/dev/disk/by-label/${SDROOT_LABEL_PFX}${name#SD}" ]; then
                readlink -f "/dev/disk/by-label/${SDROOT_LABEL_PFX}${name#SD}"
                return 0
            else
                local mmcpart="$(get_nth_mmcpartnum ${name#SD})"
                if [ -n "$mmcpart" ]; then
                    echo "$SDDEV$mmcpart"
                    return 0
                fi
            fi
            ;;
    esac

    echo "Unknown boot device: $1" >&2
    return 1
}

get_mmc_device_name()
{
    local dev="$1"
    
    local num=1
    for pnum in $SDPARTLIST; do
        if [ "$SDDEV$pnum" = "$dev" ]; then
            echo "SD$num"
            return 0
        fi
        num=$((num + 1))
    done

    return 1
}

get_device_name() {
    local dev="$1"

    if [ "${dev#$NANDDEV}" != "$dev" ]; then
        echo 'NAND'
        return 0
    elif [ "${dev#$SDDEV}" != "$dev" ]; then
        if [ -d /dev/disk/by-label ]; then
            local ret=0
            find /dev/disk/by-label -name "$SDROOT_LABEL_PFX*" | (
                while read label; do
                    if [ "$dev" = "$(readlink -f "$label")" ]; then
                        label="${label##*/}"
                        echo "SD${label#$SDROOT_LABEL_PFX}"
                        exit 0
                    fi
                done

                exit 1
            ) || ret=$?

            if [ $ret -eq 0 ]; then
                return 0
            fi
        fi

        if get_mmc_device_name "$dev"; then
            return 0
        fi
    fi

    echo 'UNKNOWN'
    return 1
}

is_nand_device() {
    case "$(get_device_name "${1:-}")" in
        NAND)
            return 0
            ;;
    esac

    return 1
}

is_mmc_device() {
    case "$(get_device_name "${1:-}")" in
        SD*)
            return 0
            ;;
    esac

    return 1
}

get_current_root_device()
{
    cat /proc/cmdline | \
        sed -e 's/[[:space:]]\+/\n/g' -e 's/\n\+/\n/g' | \
        sed -n -e 's/^root=//p' | \
        tail -1
}

get_uboot_mmc_part()
{
    if [ -n "$MMCPART_VAR" ]; then
        fw_printenv -n "$MMCPART_VAR" 2>/dev/null | \
            sed -n -e "s,^$MMCPART_MATCH\$,\\1,p"
    fi
}

get_current_bootcmd_base()
{
    local bootcmd="$(fw_printenv "$BOOT_VAR" 2> /dev/null)"
    bootcmd="$(echo "$bootcmd" | sed -e 's/^[^=]\+=//' -e 's/;$//' -e 's/^\(.*;\)[^;]*$/\1/')"
    if [ -n "$bootcmd" ]; then
        bootcmd="$bootcmd "
    fi

    echo "$bootcmd"
}

set_next_boot() {
  local name="$1"
    
  if ! is_bootname "$name"; then
    echo "Unknown boot device: $name" >&2
    return 1
  fi

  case "$name" in
      NAND)
          set_next_boot_nand
          ;;
      SD*)
          mmcpart="${name#SD}"
          set_next_boot_mmc "$mmcpart"
          ;;
      *)
          echo "Unknown boot device: $name" >&2
          return 1
          ;;
  esac
}

__fw_setenv()
{
    if [ -n "${DRY_RUN:-}" ]; then
        printf '[dry run] fw_setenv "%s" "%s"\n' "$@"
    else
        fw_setenv "$@"
    fi
}

# args: <bootname> <absolute path to fstab file>
update_fstab()
{
    local name="$1"; shift
    local fstab="$1"; shift

    is_bootname "$name" || return 1

    local sdnum=
    case "$name" in
        SD*)
            sdnum=${name#SD}
            ;;
        *)
            # No overlays for NAND partitions yet...
            touch "$fstab"
            return 0
            ;;
    esac

    for ovr in $SDOVRLIST; do
        local mpoint="${ovr%%:*}"
        local mmcpart="$(echo ${ovr#*:} | cut -d: -f $sdnum)"
        
        if [ -z "$mmcpart" ]; then
            echo "ERROR: Partition number #$sdnum is undefined for $mpoint" >&2
            return 1
        fi

        if [ -e "$fstab" ]; then
            mv -f "$fstab" "$fstab.orig"
        else
            touch "$fstab.orig"
        fi  
        
        cat "$fstab.orig" | \
        gawk -e "BEGIN { done = 0; }" \
             -e "{
                     if ( \$2 == \"$mpoint\" ) {
                         print \"$SDDEV$mmcpart\\t\" \$2 \"\\t\" \$3 \"\\t\" \$4 \"\\t\" \$5 \"\\t\" \$6;
                         done = 1;
                     } else {
                         print \$0;
                     }
                 }" \
             -e "END {
                   if ( done == 0 ) {
                     print \"$SDDEV$mmcpart\\t$mpoint\\text4\\trw,noatime,nodiratime\\t0\\t2\";
                   }
                 }" \
             >"$fstab"
    done
}

# args: <bootname>
list_overlays()
{
    local name="$1"; shift

    is_bootname "$name" || return 1

    local sdnum=
    case "$name" in
        SD*)
            sdnum=${name#SD}
            ;;
        *)
            # No overlays for NAND partitions yet...
            return 0
            ;;
    esac
    
    for ovr in $SDOVRLIST; do
        local mpoint="${ovr%%:*}"
        local mmcpart="$(echo ${ovr#*:} | cut -d: -f $sdnum)"

        [ -n "$mmcpart" ] || continue

        printf '%s\t%s\n' "$SDDEV$mmcpart" "$mpoint"
    done
}

fi #__included_boot_funcs
