#!/bin/sh

if [ -z "${__included_mount_funcs-}" ]; then
__included_mount_funcs=1

. boot-func.sh

MOUNTCONFDIR="${MOUNTCONFDIR:-/etc/rootfs-utils}"
MOUNTCONF="${MOUNTCONF:-$MOUNTCONFDIR/mount.conf}"

if [ -n "$MOUNTCONF" -a -e "$MOUNTCONF" ]; then
    . "$MOUNTCONF"
fi

ROOT_MPOINT="${ROOT_MPOINT:-/}"

MNT_NANDROOT="${MNT_NANDROOT:-/tmp/mnt/nand-root}"
MNT_SDROOT="${MNT_SDROOT:-/tmp/mnt/sd-root}"

OVERDIRS="${OVERDIRS-}"
SYSDIRS="${SYSDIRS-/proc /sys /dev /dev/pts}"
SDDIRS="${SDDIRS-/tmp/mnt/storage}"
TMPCPDIRS="${TMPCPDIRS-/etc/dhcpc}"

MKSDPART="${MKSDPART:-mkfs.ext4 -q \$dev}"

is_mounted() {
  local arg="$1"
  if is_bootname "$arg"; then
      local dev="$(get_root_device "$arg")"
      mount -l | grep -q "^$dev[[:space:]]"
  else
      mount -l | grep -qw "$arg"
  fi
}

get_root_mpoint()
{
    local name="$1"

    case "$name" in
        NAND)
            echo "$MNT_NANDROOT"
            ;;
        SD*)
            echo "$MNT_SDROOT${name#SD}"
            ;;
        *)
            echo "Unknown boot device: $1" >&2
            return 1
            ;;
    esac
}

get_mpoint_name() {
    local mpoint="$1"

    case "$mpoint" in
        $MNT_NANDROOT)
            echo 'NAND'
            ;;
        ${MNT_SDROOT}*)
            echo "SD${mpoint#$MNT_SDROOT}"
            ;;
        *)
            echo 'UNKNOWN'
            return 1
            ;;
    esac
}

is_nand_mpoint() {
    [ "$(get_mpoint_name "${1:-}")" = 'NAND' ]
}

is_sd_mpoint() {
    case "$(get_mpoint_name "${1:-}")" in
        SD*)
            return 0
            ;;
    esac
    
    return 1
}

attach_nand() {
    local dev="${1:-$(get_root_device NAND)}"
    if [ ! -e /dev/ubi0 ]; then
        ubiattach /dev/ubi_ctrl -m $ROOT_NAND_NUM
    fi
}

detach_nand() {
    local dev="${1:-$(get_root_device NAND)}"
    if [ -e /dev/ubi0 ]; then
        ubidetach /dev/ubi_ctrl -m $ROOT_NAND_NUM
    fi
}

# args: <path to /etc/fstab file>
# output: /dev \t /mpoint
list_automounts()
{
    local fstab="${1:-/etc/fstab}"

    [ -e "$fstab" ] || return 0
    
    gawk -e '/noauto|^\s*#|^$/ { next; }' -e '/^\/|^[Uu][Uu][Ii][Dd]=/ { if ($2 ~ /^\//) { print $1 "\t" $2; } }' "$fstab"
}

# args: <chroot> <dev> <mpoint>
mount_overdir()
{
    local rootdir="$1"; shift
    local dev="$1"; shift
    local mpoint="$1"; shift

    ! is_mounted "${rootdir%/}/${mpoint#/}" || return 0
    [ -d "${rootdir%/}/${mpoint#/}" ] || mkdir -p "${rootdir%/}/${mpoint#/}"
    mount "$@" "$dev" "${rootdir%/}/${mpoint#/}"
}

# depends: $OVERDIRS
# args: <path to chroot> [ mount args ... ]
mount_overdirs()
{
    [ -n "$OVERDIRS" ] || return 0
    
    local rootdir="$1"; shift

    list_automounts "${rootdir%/}/etc/fstab" | \
        while read dev mpoint; do
            echo "$OVERDIRS" | grep -qw "$mpoint" || continue
            mount_overdir "$rootdir" "$dev" "$mpoint"
        done
}

# args: <chroot> <mpoint>
umount_overdir()
{
    local rootdir="$1"; shift
    local mpoint="$1"; shift

    is_mounted "${rootdir%/}/${mpoint#/}" || return 0
    umount "$@" "${rootdir%/}/${mpoint#/}"
}

# args: <path to chroot> [ umount args ... ]
umount_overdirs()
{
    [ -n "$OVERDIRS" ] || return 0
    
    local rootdir="$1"; shift

    mount -l | gawk -e "{
                 if ( \$3 ~ \"^$rootdir/\" ) {
                     print \$3;
                 }
                        }" | \
        while read mpoint; do
            umount_overdir "$rootdir" "${mpoint#$rootdir}"
        done
}

mountfs()
{
  local dev="$1"; shift
  if is_bootname "$dev"; then
      local mpoint="$(get_root_mpoint "$dev")"
      dev="$(get_root_device "$dev")"
  else
      local mpoint="$1"; shift
  fi
  
  if ! is_mounted "$dev";  then
    mkdir -p "$mpoint"
    case "$(get_device_name "$dev")" in
        NAND)
            attach_nand "$dev"
            mount -t ubifs ubi0:rootfs "$@" "$mpoint"
            ;;
        SD*)
            mount "$@" "$dev" "$mpoint"
            ;;
        *)
            echo "Unsupported device: $dev" >&2
            return 1
            ;;
    esac
  fi
  
  if is_mounted "$mpoint";  then
      mount_overdirs "$mpoint" "$@"
  fi
}

bind_dir() {
    local dir="$1"
    local mpoint="$2"

    if [ -d "$dir" ]; then
        if ! is_mounted "${mpoint%/}/${dir#/}"; then
            mkdir -p "${mpoint%/}/${dir#/}"
            mount -o bind "$dir" "${mpoint%/}/${dir#/}"
        fi
    fi
}

bind_dirs() {
    local mpoint="$1"; shift

    if [ $# -gt 0 ]; then
        for dir in "$@"; do
            bind_dir "$dir" "$mpoint"
        done
    fi
}

bind_tmp() {
    local mpoint="$1"
    mpoint="${mpoint%/}"
    tmpdir="/tmp/.rootdir/${mpoint#/}"

    mkdir -p "$tmpdir"
    mkdir -p "$mpoint/tmp"

    is_mounted "$mpoint/tmp" || \
        mount -o bind "$tmpdir" "$mpoint/tmp"

    if [ -n "$TMPCPDIRS" ]; then
        for dir in $TMPCPDIRS; do
            dir="${dir#/}"
            mkdir -p "$mpoint/tmp/${dir%/}"
            cp -af "/tmp/${dir%/}/." "$mpoint/tmp/${dir%/}/"
        done
    fi
}

ubind_dir() {
    local dir="$1"

    if [ -d "$dir" ] && is_mounted "$dir"; then
        umount "$dir"
    fi
}

ubind_dirs() {
    local mpoint="$1"; shift

    if [ $# -gt 0 ]; then
        for dir in "$@"; do
            ubind_dir "${mpoint%/}/${dir#/}"
        done
    fi
}

ubind_tmp() {
    local mpoint="$1"
    mpoint="${mpoint%/}"
    tmpdir="/tmp/.rootdir/${mpoint#/}"

    is_mounted "$mpoint/tmp" || return 0
    
    umount "$mpoint/tmp"
    rm -rf "$tmpdir"
}

mountrootfs() {
    mountfs "$@"

    local dev="$1"; shift
    if is_bootname "$dev"; then
        local mpoint="$(get_root_mpoint "$dev")"
        dev="$(get_root_device "$dev")"
    else
        local mpoint="$1"; shift
    fi

    bind_dirs "$mpoint" $SYSDIRS $SDDIRS
    bind_tmp "$mpoint"
}

umountfs()
{
  local dev="$1"; shift
  if is_bootname "$dev"; then
      local mpoint="$(get_root_mpoint "$dev")"
      dev="$(get_root_device "$dev")"
  else
      local mpoint="${1:-}"
      [ $# -eq 0 ] || shift
  fi
  
  if [ -d "$dev" ]; then
      mpoint="$dev"
      if is_mounted "$mpoint";  then
          umount_overdirs "$mpoint" "$@"
          umount "$@" "$mpoint"
          case "$(get_mpoint_name "$mpoint")" in
              NAND)
                  detach_nand
                  ;;
          esac
      fi
  else
      if is_mounted "$dev";  then
          umount_overdirs "$mpoint" "$@"
          umount "$@" "$mpoint"
          case "$(get_device_name "$dev")" in
              NAND)
                  detach_nand "$dev"
                  ;;
              SD*)
                  ;;
              *)
                  echo "Unsupported device: $dev" >&2
                  return 1
                  ;;
          esac
      fi
  fi
}

umountrootfs() {
    local dev="$1"
    if is_bootname "$dev"; then
        local mpoint="$(get_root_mpoint "$dev")"
        dev="$(get_root_device "$dev")"
    else
        local mpoint="$2"
    fi

    ubind_tmp "$mpoint"

    if [ -n "$SYSDIRS $SDDIRS" ]; then
        ubind_dirs "$mpoint" $(echo $SYSDIRS $SDDIRS | tac -r -s '\s+')
    fi
    
    umountfs "$@"
}

remountfs() {
  local mpoint="$1"; shift
  ! is_bootname "$mpoint" || mpoint="$(get_root_mpoint "$mpoint")"
  mount -o remount "$@" "$mpoint"
  mount_overdirs "$mpoint" -o remount "$@"
}

erase_sdpart()
{
    local dev="$1"; shift
    if is_bootname "$dev"; then
        dev="$(get_root_device "$dev")"
    fi

    if is_mounted "$dev"; then
        echo "ERROR: $dev is mounted" >&2
        return 1
    fi

    eval "$MKSDPART"
}

fi #__included_mount_funcs
