import dateutil.parser
from datetime import datetime
from txtstruct import *
import reportlab.graphics.shapes as rlgraphsh
import reportlab.graphics.charts.barcharts as rlgraphch
import reportlab.platypus
import os.path

class TestInfo:

  started  = None
  finished = None

  def __init__(self, started, finished):
    self.started  = started
    self.finished = finished

  @staticmethod
  def from_dict(dct):
    tstart  = dateutil.parser.parse(dct['started'])
    tfinish = (dateutil.parser.parse(dct['finished'])
      if not dct['finished'] is None else None)
    return TestInfo(tstart, tfinish)

  @staticmethod
  def format_time(datetime):
    """Formats data into form applicable for usage in reports"""
    return datetime.strftime('%d %b %Y %H:%M:%S (UTC %z)')

  def to_report(self, strings, level = 3):
    title = Paragraph(strings['common']['various'], level)
    info = Dictionary([
      (strings['common']['started'],
         TestInfo.format_time(self.started)),
      (strings['common']['finished'], strings['common']['no_fin']
         if self.finished is None else
         TestInfo.format_time(self.finished))
    ])
    return Group([title, info])

class PortInfo:

  rate_mbps = None
  rate_set  = None
  mac     = None
  enabled = None

  def __init__(self, rate_mbps, rate_set, mac, enabled):
    self.rate_mbps = rate_mbps
    self.rate_set  = rate_set
    self.mac     = mac
    self.enabled = enabled

  @staticmethod
  def from_dict(dct):
    if dct is None:
      return None
    return PortInfo(dct['rate_mbps'], dct['rate_set_mbps'],
      dct['mac'], dct['enabled'])

  def to_report(self, strings, level = None):
    return Dictionary([
      (strings['common']['link_rate'],
         strings['common']['null'] if self.rate_mbps is None else
         ' '.join([str(self.rate_mbps), strings['units']['mbps']])),
      (strings['common']['rate_set'],
         strings['common']['autoneg'] if self.rate_set is None else
         ' '.join([str(self.rate_set), strings['units']['mbps']])),
      (strings['common']['mac'], self.mac)
    ]) if self.enabled else None

class PortsInfo:

  a = None
  b = None

  def __init__(self, a, b):
    self.a = a
    self.b = b

  @staticmethod
  def from_dict(dct):
    a = PortInfo.from_dict(dct['a'])
    b = PortInfo.from_dict(dct['b'])
    return PortsInfo(a, b)

  def to_report(self, strings, level = 3):
    res = []
    res += [Paragraph(strings['common']['ports'], level)]
    a = self.a.to_report(strings)
    b = self.b.to_report(strings)
    res += [Dictionary([
      (strings['common']['port']['a'],
         strings['common']['disabled' if a is None else 'enabled']),
      (strings['common']['port']['b'],
         strings['common']['disabled' if b is None else 'enabled']),
    ])]

    if not a is None:
      res += [Paragraph(strings['common']['port']['a'], level + 1), a]

    if not b is None:
      res += [Paragraph(strings['common']['port']['b'], level + 1), b]

    return Group(res)

class HeaderInfo:

  ipdst = None
  ipsrc = None
  macdst = None
  macsrc = None
  udpdst = None
  udpsrc = None
  tos = None
  mpls = None
  vlan = None
  precedence = None

  def __init__(self, ipdst, ipsrc, macdst, macsrc, udpdst, udpsrc):
    self.ipdst  = ipdst
    self.ipsrc  = ipsrc
    self.macdst = macdst
    self.macsrc = macsrc
    self.udpdst = udpdst
    self.udpsrc = udpsrc

  @staticmethod
  def from_dict(dct):
    if dct is None:
      return None
    return HeaderInfo(dct['dst']['ip'], dct['src']['ip'],
        dct['dst']['mac'], dct['src']['mac'],
        dct['dst']['udp'], dct['src']['udp'])

  def to_report(self, strings, level = 3):
    title = Paragraph(strings['common']['header'], level)

    tit_src = Paragraph(strings['common']['src'], level + 1)
    src = Dictionary([
      (strings['common']['ip'],       self.ipsrc),
      (strings['common']['mac'],     self.macsrc),
      (strings['common']['udpport'], self.udpsrc)
    ])

    tit_dst = Paragraph(strings['common']['dst'], level + 1)
    dst = Dictionary([
      (strings['common']['ip'],       self.ipdst),
      (strings['common']['mac'],     self.macdst),
      (strings['common']['udpport'], self.udpdst)
    ])

    return Group([title, tit_src, src, tit_dst, dst])

class TopologyInfo:

  remote_ip = None
  rx_port   = None
  tx_port   = None

  def __init__(self, remote_ip, rx_port, tx_port):
    self.remote_ip = remote_ip
    self.rx_port   = rx_port
    self.tx_port   = tx_port

  @staticmethod
  def from_dict(dct):
    if dct is None:
      return None
    return TopologyInfo(dct['remote_ip'], dct['rx'], dct['tx'])

  def to_report(self, strings, level = 2):
    title = Paragraph(strings['common']['topology'], level)

    dct = Dictionary([
      (strings['common']['remote_ip'], self.remote_ip),
      (strings['common']['tx_port'],
         strings['common']['port'][self.tx_port]),
      (strings['common']['rx_port'],
         strings['common']['port'][self.rx_port])
    ])

    return Group([title, dct])

class RateInfo:

  rate_type = None
  value = None

  def __init__(self, rate_type, value):
    self.rate_type = rate_type
    self.value = value

  @staticmethod
  def from_dict(dct):
    return RateInfo(dct['type'], dct['value'])

  def to_string(self, strings):
    return str(self.value) + ('%' if self.rate_type == 'percent'
      else ' ' + strings['units'][self.rate_type])

class MtkTest:

  info     = None
  ports    = None
  settings = None
  results  = None

  def __init__(self, info, ports, settings, results):
    self.info     = info
    self.ports    = ports
    self.settings = settings
    self.results  = results

  @classmethod
  def settings_hook(cls, dct):
    return cls.Settings.from_dict(dct['settings'])

  @classmethod
  def results_hook(cls, dct):
    return cls.Results.from_dict(dct['results'])

  @classmethod
  def from_dict(cls, dct):
    return cls(TestInfo.from_dict(dct['info']),
      None if not 'ports' in dct or dct['ports'] is None else
        PortsInfo.from_dict(dct['ports']),
      cls.settings_hook(dct),
      cls.results_hook(dct))

  def _to_report(self, title, strings, level = 1):
    return Group([Paragraph(title, level)] +
      map(lambda x : x.to_report(strings, level + 1),
      filter(None, [self.info, self.ports, self.settings,
        self.results]))
    )

class Rfc2544Test(MtkTest):

  class Results:

    tests = dict (
      btb = None,
      latency = None,
      throughput = None,
      loss = None
    )

    class Meta:

      measurements = None

      def __init__(self, measurements):
        self.measurements = measurements

      @classmethod
      def from_arr(cl, arr):
        return cl(map(cl.Measurement.from_dict, arr))

      def to_report(self, strings, level = 3):
        tbl = Table([self.table_header(strings)] +
          map(lambda x : x.to_table_row(strings), self.measurements))
        return tbl

    class Btb(Meta):

      def table_header(self, strings):
        return [
          strings['common']['frame'],
          strings['common']['rate'],
          strings['rfc2544']['btb']['frame_count'],
          strings['rfc2544']['trial_duration'],
          strings['common']['status']
        ]

      class Measurement:

        frame = None
        rate_pr = None
        status = None
        frame_count = None
        trial_duration = None

        def __init__(self, frame, rate_pr, status, frame_count,
          trial_duration):
          self.frame = frame
          self.rate_pr = rate_pr
          self.status = status
          self.frame_count = frame_count
          self.trial_duration = trial_duration

        @staticmethod
        def from_dict(dct):
          return Rfc2544Test.Results.Btb.Measurement(
            dct['frame'], dct['rate_%'], dct['status'],
            dct['trial_count'], dct['trial_duration']
          )

        def to_table_row(self, strings):
          return [
            TextUnit(str(self.frame) + ' ' + strings['units']['B']),
            TextUnit(str(self.rate_pr) + '%'),
            TextUnit(self.frame_count),
            TextUnit(str(self.trial_duration) + ' ' +
              strings['units']['s']),
            TextUnit(strings['statuses'][self.status.lower()])
          ]

      def to_report(self, strings, level = 3):
        tbl = Table([self.table_header(strings)] +
          map(lambda x : x.to_table_row(strings), self.measurements))

        data = '\n'.join(map(lambda x : str(x.frame) + ' ' +
          str(x.trial_duration), self.measurements))
        axis = (
          """%s (%s)""" % (strings['common']['frame'],
            strings['units']['B']),
          """%s (%s)""" % (strings['rfc2544']['btb']['time'],
            strings['units']['s'])
        )
        gr = Graphic(data, axis, os.path.dirname(__file__) +
          '/plot/thr_btb.gnu')

        return Group([tbl, gr])

    class Latency(Meta):

      class Measurement:

        delay = None
        frame = None
        rate  = None
        status = None

        def __init__(self, delay, frame, rate, status):
          self.delay = delay
          self.frame = frame
          self.rate  = rate
          self.status = status

        @classmethod
        def from_dict(cl, dct):
          return cl(dct['delay_ms'], dct['frame'], dct['rate_%'],
            dct['status'])

        def to_table_row(self, strings):
          return [
            str(self.frame) + ' ' + strings['units']['B'],
            str(self.delay) + ' ' + strings['units']['ms'],
            str(self.rate) + '%',
            strings['statuses'][self.status.lower()]
          ]

      def table_header(self, strings):
        return [
          strings['common']['frame'],
          strings['rfc2544']['latency']['delay_hdr'],
          strings['common']['rate'],
          strings['common']['status']
        ]

      def to_report(self, strings, level = 3):
        tbl = Table([self.table_header(strings)] +
          map(lambda x : x.to_table_row(strings), self.measurements))

        data = '\n'.join(map(lambda x : str(x.frame) + ' ' +
          str(x.delay), self.measurements))

        axis = (
          """%s (%s)""" % (strings['common']['frame'],
            strings['units']['B']),
          """%s (%s)""" % (strings['rfc2544']['latency']['name'],
            strings['units']['ms'])
        )
        gr = Graphic(data, axis, os.path.dirname(__file__) +
          '/plot/latency.gnu')

        return Group([tbl, gr])

    class FrameLoss(Meta):

      class Measurement:

        frame = None
        loss = None
        rate = None
        loss_l1 = None
        loss_l2 = None
        loss_l3 = None
        loss_l4 = None

        def __init__(self, frame, rate, loss, loss_l1, loss_l2,
            loss_l3, loss_l4):
          self.frame = frame
          self.rate  = rate
          self.loss  = loss
          self.loss_l1 = loss_l1
          self.loss_l2 = loss_l2
          self.loss_l3 = loss_l3
          self.loss_l4 = loss_l4

        @classmethod
        def from_dict(cl, dct):
          return cl(dct['frame'], dct['rate_%'], dct['loss_%'],
            dct['loss_mbps_l1'], dct['loss_mbps_l2'],
            dct['loss_mbps_l3'], dct['loss_mbps_l4'])

        def to_table_row(self, strings):
          return [
            str(self.frame) + ' ' + strings['units']['B'],
            str(self.rate) + '%',
            str(self.loss) + '%',
            str(self.loss_l1) + ' ' + strings['units']['mbps'],
            str(self.loss_l2) + ' ' + strings['units']['mbps'],
            str(self.loss_l3) + ' ' + strings['units']['mbps'],
            str(self.loss_l4) + ' ' + strings['units']['mbps']
          ]

      def to_report(self, strings, level = 3):
        tbl = Table([self.table_header(strings)] +
          map(lambda x : x.to_table_row(strings), self.measurements))

        adata = []
        for frame in sorted(set(map(lambda x : x.frame,
            self.measurements))):
          curf = filter(lambda x : x.frame == frame,
            self.measurements)
          cdata = '\n'.join(map(lambda x : str(x.rate) + ' ' +
            str(x.loss), sorted(curf, key = lambda x : x.rate)))
          cdata = ('"' + str(frame) + ' ' + strings['units']['btes']
            + '"\n' + cdata)
          adata.append(cdata)

        data = '\n\n\n'.join(adata)

        axis = (
          strings['common']['rate'] + ' (%)',
          strings['rfc2544']['loss']['name'] + ' (%)'
        )
        gr = Graphic(data, axis, os.path.dirname(__file__) +
          '/plot/frameloss.gnu')

        return Group([tbl, gr])

      def table_header(self, strings):
        return [
          strings['common']['frame'],
          strings['common']['rate'],
          strings['rfc2544']['loss']['loss_hdr'],
          strings['rfc2544']['loss']['loss_l1_hdr'],
          strings['rfc2544']['loss']['loss_l2_hdr'],
          strings['rfc2544']['loss']['loss_l3_hdr'],
          strings['rfc2544']['loss']['loss_l4_hdr']
        ]

    class Throughput(Meta):

      measurements = None
      trials = None

      class Trial:

        loss = None
        pkts_rx = None
        pkts_tx = None
        rate = None
        rate_l1 = None
        status = None
        trial_num = None

        def __init__(self, loss, pkts_rx, pkts_tx, rate, rate_l1,
            status, trial_num):
          self.loss = loss
          self.pkts_rx = pkts_rx
          self.pkts_tx = pkts_tx
          self.rate = rate
          self.rate_l1 = rate_l1
          self.status = status
          self.trial_num = trial_num

        @classmethod
        def from_dict(cl, dct):
          return cl(dct['loss'], dct['pkts_rx'], dct['pkts_tx'],
            dct['rate_%'], dct['rate_mbps_l1'], dct['status'],
            dct['trial_num'])

        def to_table_row(self, strings):
          return [ self.loss, self.pkts_rx, self.pkts_tx, self.rate,
            self.rate_l1, self.trial_num,
            strings['statuses'][self.status.lower()] ]

      class Measurement:

        rate = None
        l1 = None
        l2 = None
        l3 = None
        l4 = None
        frame = None
        frm_per_second = None
        status = None

        def __init__(self, frame, frm_per_second, rate, l1, l2, l3,
            l4, status):
          self.frame = frame
          self.frm_per_second = frm_per_second
          self.rate = rate
          self.l1 = l1
          self.l2 = l2
          self.l3 = l3
          self.l4 = l4
          self.status = status

        @classmethod
        def from_dict(cl, dct):
          return cl(dct['frame'], dct['frm_per_second'],
            dct['rate_%'], dct['rate_mbps_l1'],
            dct['rate_mbps_l2'], dct['rate_mbps_l3'],
            dct['rate_mbps_l4'], dct['status'])

        def to_table_row(self, strings):
          return [
            str(self.frame) + ' ' + strings['units']['B'],
            self.frm_per_second,
            str(self.rate) + '%',
            str(self.l1) + ' ' + strings['units']['mbps'],
            str(self.l2) + ' ' + strings['units']['mbps'],
            str(self.l3) + ' ' + strings['units']['mbps'],
            str(self.l4) + ' ' + strings['units']['mbps'],
            strings['statuses'][self.status.lower()]
          ]

      def table_header(self, strings):
        return ([
          strings['common']['frame'],
          strings['rfc2544']['throughput']['frm_per_sec'],
          strings['common']['rate'],
          strings['rfc2544']['throughput']['rate_hdrs']['l1'],
          strings['rfc2544']['throughput']['rate_hdrs']['l2'],
          strings['rfc2544']['throughput']['rate_hdrs']['l3'],
          strings['rfc2544']['throughput']['rate_hdrs']['l4'],
          strings['common']['status']
        ])

      def to_report(self, strings, level = 3):
        tbl = Table([self.table_header(strings)] +
          map(lambda x : x.to_table_row(strings), self.measurements))

        data = '\n'.join(map(lambda x : str(x.frame) + ' ' +
          str(x.l1), self.measurements))
        axis = (
          """%s (%s)""" % (strings['common']['frame'],
            strings['units']['B']),
          """%s (%s)""" %
            (strings['rfc2544']['throughput']['rate_hdrs']['l1'],
            strings['units']['mbps'])
        )
        gr = Graphic(data, axis, os.path.dirname(__file__) +
          '/plot/thr_btb.gnu')

        return Group([tbl, gr])

    def __init__(self, tests):
      self.tests['btb'] = tests['btb']
      self.tests['latency'] = tests['latency']
      self.tests['throughput'] = tests['throughput']
      self.tests['loss'] = tests['loss']

    @staticmethod
    def from_dict(dct):
      btb = Rfc2544Test.Results.Btb.from_arr(dct['backtoback'])
      latency = Rfc2544Test.Results.Latency.from_arr(dct['latency'])
      throughput = Rfc2544Test.Results.Throughput.from_arr(
        dct['throughput'])
      loss = Rfc2544Test.Results.FrameLoss.from_arr(dct['frameloss'])
      return Rfc2544Test.Results(dict(
        btb = btb,
        latency = latency,
        throughput = throughput,
        loss = loss))

    def to_report(self, strings, level = 2):
      return Group(map(lambda x : self.tests[x].to_report(strings,
        level + 1), ['throughput', 'loss', 'btb', 'latency']))

  class Settings:

    topology = None
    frames = None
    header = None
    learn  = None
    wait   = None
    tests = dict (
      latency = None,
      btb = None,
      throughput = None,
      loss = None
    )

    class Rate:

      value = None
      units = None

      def __init__(self, value, units):
        self.value = str(value)
        self.units = units.lower()

      @staticmethod
      def from_dict(dct):
        return Rfc2544Test.Settings.Rate(dct['value'], dct['units'])

      def to_report(self, strings, level = None):
        if self.units == '%':
          s = self.value + '%'
        elif self.units == 'mbps':
          s = self.value + ' ' + strings['units']['mbps']
        elif self.units == 'kbps':
          s = self.value + ' ' + strings['units']['kbps']
        return s

    class Rates:

      rates = None

      def __init__(self, rates):
        self.rates = rates

      @staticmethod
      def from_arr(arr):
        return Rfc2544Test.Settings.Rates(
          map(Rfc2544Test.Settings.Rate.from_dict, arr))

      def to_report(self, strings, level = 3):
        rt = Paragraph(strings['rfc2544']['rates'], level)

        rates = map(lambda x : x.to_report(strings, level),
          self.rates)
        r = map(lambda i : (str(i + 1), rates[i]),
          range(0, len(rates)))

        return Group([rt, r])

    class BtbOrLatency:

      rates = None
      rates_src = None
      trial_count = None
      trial_duration = None
      enabled = None

      def __init__(self, rates, rates_src, trial_count,
        trial_duration, enabled):
        self.rates = rates
        self.rates_src = rates_src
        self.trial_count = trial_count
        self.trial_duration = trial_duration
        self.enabled = enabled

      @staticmethod
      def from_dict(dct):
        rates = Rfc2544Test.Settings.Rates.from_arr(dct['rates'])
        return Rfc2544Test.Settings.BtbOrLatency(rates,
          dct['rates_src'], dct['trial_count'],
          dct['trial_duration'], dct['enabled'])

      def to_report(self, strings, level = 3):

        grp = []

        settings_grp = []
        if self.rates_src == 'throughput':
          settings_grp.append(
            (strings['rfc2544']['rates_src'],
             strings['rfc2544']['rate_src_thr']))

        settings_grp.extend([
          (strings['rfc2544']['trial_count'],    self.trial_count),
          (strings['rfc2544']['trial_duration'],
             str(self.trial_duration) + ' ' + strings['units']['s'])
        ])

        settings = Dictionary(settings_grp)
        grp.append(settings)

        if self.rates_src == 'manual':
          grp.extend(self.rates.to_report(strings, level + 1))

        return Group(grp)

    class Throughput:

      maxrate = None
      resolution = None
      threshold = None
      trial_duration = None
      enabled = None

      def __init__(self, maxrate, resolution, threshold,
        trial_duration, enabled):
        self.maxrate = maxrate
        self.resolution = resolution
        self.threshold = threshold
        self.trial_duration = trial_duration
        self.enabled = enabled

      @staticmethod
      def from_dict(dct):
        return Rfc2544Test.Settings.Throughput(
          Rfc2544Test.Settings.Rate.from_dict(dct['maxrate']),
          dct['resolution'], dct['threshold'], dct['trial_duration'],
          dct['enabled'])

      def to_report(self, strings, level = 3):
        settings = Dictionary([
          (strings['rfc2544']['throughput']['maxrate'],
             self.maxrate.to_report(strings, level + 1)),
          (strings['rfc2544']['throughput']['resolution'],
             str(self.resolution) + '%'),
          (strings['rfc2544']['throughput']['threshold'],
             str(self.threshold) + '%'),
          (strings['rfc2544']['trial_duration'],
             str(self.trial_duration) + ' ' + strings['units']['s'])
        ])

        return settings

    class FrameLoss:

      rates_start = None
      rates_end = None
      steps = None
      trial_duration = None
      enabled = None

      def __init__(self, rates_start, rates_end, steps,
        trial_duration, enabled):
        self.rates_start = rates_start
        self.rates_end = rates_end
        self.steps = steps
        self.trial_duration = trial_duration
        self.enabled = enabled

      @classmethod
      def from_dict(cls, dct):
        rates_start = Rfc2544Test.Settings.Rate.from_dict(
          dct['rates_start'])
        rates_end   = Rfc2544Test.Settings.Rate.from_dict(
          dct['rates_end'])
        return cls(rates_start, rates_end, dct['steps'],
          dct['trial_duration'], dct['enabled'])

      def to_report(self, strings, level = 3):
        settings = Dictionary([
          (strings['rfc2544']['loss']['rates_start'],
             self.rates_start.to_report(strings, level + 1)),
          (strings['rfc2544']['loss']['rates_end'],
             self.rates_end.to_report(strings, level + 1)),
          (strings['rfc2544']['loss']['steps'], self.steps),
          (strings['rfc2544']['trial_duration'],
             str(self.trial_duration) + ' ' + strings['units']['s'])
        ])
        return settings

    def __init__(self, topology, frames, header, learn, wait, tests):
      self.topology = topology
      self.frames = frames
      self.header = header
      self.learn  = learn
      self.wait   = wait
      for name in ['latency', 'btb', 'throughput', 'loss']:
        self.tests[name] = tests[name]

    @staticmethod
    def from_dict(dct):
      topology = TopologyInfo.from_dict(dct['topology'])
      frames = filter(None, map(lambda x : x['size'] if x['enabled']
        else None, dct['frames']))
      header = HeaderInfo.from_dict(dct['header'])
      learn = dct['config']['learn_ms']
      wait  = dct['config']['wait_ms']
      tests = dict(
        latency = Rfc2544Test.Settings.BtbOrLatency.from_dict(
          dct['config']['latency']),
        btb = Rfc2544Test.Settings.BtbOrLatency.from_dict(
          dct['config']['backtoback']),
        throughput = Rfc2544Test.Settings.Throughput.from_dict(
          dct['config']['throughput']),
        loss = Rfc2544Test.Settings.FrameLoss.from_dict(
          dct['config']['frameloss'])
      )
      return Rfc2544Test.Settings(topology, frames,
        header, learn, wait, tests)

    def to_report(self, strings, level = 3):

      grp = []

      grp.append(Paragraph(strings['common']['settings'], level))

      settings = Dictionary( [
        ( strings['common']['frames'], ', '.join(map(str,
            self.frames)) ),
        ( strings['rfc2544']['learn'], str(self.learn) +
            ' ' + strings['units']['ms']),
        ( strings['rfc2544']['wait'],  str(self.wait)  +
            ' ' + strings['units']['ms'])
      ] )

      grp.append(settings)

      header = self.header.to_report(strings, level + 1)
      grp.append(header)

      if not self.topology is None:
        topology = self.topology.to_report(strings, level + 1)
        grp.append(topology)

      return Group(grp)

  def to_report(self, strings, level = 1):
    grp = []

    title = Paragraph(strings['rfc2544']['name'], level)
    grp.append(title)

    for t in filter(None, [self.info, self.ports, self.settings]):
      grp.append(t.to_report(strings, level + 1))

    for test in filter(lambda x : self.settings.tests[x].enabled,
        ['throughput', 'latency', 'loss', 'btb']):
      ttitle = Paragraph(strings['rfc2544'][test]['name'], level + 1)
      setttitle = Paragraph(strings['common']['settings'], level + 2)
      sett = self.settings.tests[test].to_report(strings,  level + 2)
      restitle = Paragraph(strings['common']['results'],   level + 2)
      tres = self.results.tests[test].to_report(strings,   level + 2)
      grp.extend([ttitle, setttitle, sett, restitle, tres])

    return Group(grp)

class BerTest(MtkTest):

  class Results:

    ber = None
    bits = None
    ebits = None
    elapsed = None
    los = None
    los_pr = None
    lss = None
    lss_pr = None
    remaining = None

    def __init__(self, ber, bits, ebits, elapsed, los, los_pr, lss,
        lss_pr, remaining):
      self.ber = ber
      self.bits = bits
      self.ebits = ebits
      self.elapsed = elapsed
      self.los = los
      self.los_pr = los_pr
      self.lss = lss
      self.lss_pr = lss_pr
      self.remaining = remaining

    @staticmethod
    def from_dict(dct):
      return BerTest.Results(dct['ber'], dct['bits'],
          dct['ebits'], dct['elapsed'], dct['los'], dct['los_%'],
          dct['lss'], dct['lss_%'], dct['remaining'])

    def to_report(self, strings, level = 2):

      def unsync(val):
        return strings['bert']['unsync'] if val is None else val

      title = Paragraph(strings['common']['results'], level)
      results = Dictionary( [
        ( strings['bert']['ber'],    unsync(self.ber)    ),
        ( strings['bert']['bits'],   self.bits           ),
        ( strings['bert']['ebits'],  self.ebits          ),
        ( strings['bert']['lss'],    self.lss            ),
        ( strings['bert']['lss_pr'], unsync(self.lss_pr) ),
        ( strings['bert']['los'],    self.los            ),
        ( strings['bert']['los_pr'], unsync(self.los_pr) )
      ] )
      return Group( [ title, results ] )

  class Settings:

    class Pattern:
      ptype = None
      param = None
      def __init__(self, ptype, param):
        self.ptype = ptype.lower()
        self.param = param
      @staticmethod
      def from_dict(dct):
        return BerTest.Settings.Pattern(
          dct['type'], dct['param'])
      def to_report(self, strings, level = None):
        s = strings['bert']['pattern']['types'][self.ptype]
        if (self.ptype == 'user') or (self.ptype == 'prbs'):
          s += (strings['bert']['pattern'][self.ptype]['param_pre']
                + str(self.param) +
                strings['bert']['pattern'][self.ptype]['param_post'])
        return TextUnit(s)

    header = None
    topology = None
    duration = None
    frame = None
    level = None
    pattern = None
    rate = None

    def __init__(self, header, topology, duration, frame, level,
        pattern, rate):
      self.header   = header
      self.topology = topology
      self.duration = duration
      self.frame    = frame
      self.level    = level
      self.pattern  = pattern
      self.rate     = rate

    @staticmethod
    def from_dict(dct):
      return BerTest.Settings(
        HeaderInfo.from_dict(dct['header']),
        TopologyInfo.from_dict(dct['topology']),
        dct['config']['duration'],
        dct['config']['frame'],
        dct['config']['level'],
        BerTest.Settings.Pattern.from_dict(
          dct['config']['pattern']),
        RateInfo.from_dict(dct['config']['rate'])
	)

    def to_report(self, strings, level = 2):

      grp = []

      grp.append(Paragraph(strings['common']['settings'], level))

      pattern = self.pattern.to_report(strings)
      settings = Dictionary( [
        ( strings['bert']['pattern']['pattern'], pattern),
        ( strings['common']['duration'], str(self.duration) +
            ' ' + strings['units']['s']),
        ( strings['common']['frame'], str(self.frame) + ' ' +
            strings['units']['B']),
        ( strings['common']['level'],    self.level ),
        ( strings['common']['rate'], self.rate.to_string(strings) )
      ] )

      grp.append(settings)

      grp.append(self.header.to_report(strings, level + 1))

      if not self.topology is None:
        grp.append(self.topology.to_report(strings, level + 1))

      return Group(grp)

  def to_report(self, strings, level = 1):
    return self._to_report(strings['bert']['name'],
      strings, level)

class Y1564Test(MtkTest):

  streams  = None
  settings = None

  class Streams:

    results = None
    settings = None

    class Results:

      configuration = None
      performance = None
  
      class Configuration:
  
        measurements = None
  
        class Measurement:
      
          fdv = None
          flr = None
          ftd = None
          ir  = None
          status = None
    
          def __init__(self, fdv, flr, ftd, ir, status):
            self.fdv = fdv
            self.flr = flr
            self.ftd = ftd
            self.ir  = ir
            self.status = status
    
          @classmethod
          def from_dict(cl, dct):
            return cl(dct['fdv_ms'], dct['flr'], dct['ftd_ms'],
              dct['ir'], dct['status'])
    
          def to_table_row(self, strings):
            return [strings['statuses'][self.status.lower()],
              self.ir, self.flr, self.ftd, self.fdv]
  
        def __init__(self, measurements):
          self.measurements = measurements
  
        @classmethod
        def from_dict(cls, dct):
          g = []
          for i in range(len(dct['cir'])):
            c = dct['cir'][i]
            if not c is None:
              g.append( ('CIR' + ' ' + str(i+1),
                cls.Measurement.from_dict(c)) )
  
          for t in ['eir', 'tp']:
            c = dct[t]
            if not c is None:
              g.append( (t.upper(),
                cls.Measurement.from_dict(c)) )
  
          return cls(g)
  
        def to_report(self, strings, level = 4):
          return Table(
            [[strings['y1564']['stage'], strings['common']['status'],
              strings['y1564']['ir'], strings['y1564']['loss'],
              strings['y1564']['ftd'], strings['y1564']['fdv']]] +
            map(lambda x : [x[0]] + x[1].to_table_row(strings),
            self.measurements))
  
      class Performance:
  
        class NumberRange:
  
          mn  = None
          avg = None
          mx  = None
  
          def __init__(self, mn, avg, mx):
            self.mn  = mn
            self.avg = avg
            self.mx  = mx
  
          @classmethod
          def from_dict(cl, dct):
            return cl(dct['min'], dct['avg'], dct['max'])
  
          def __str__(self):
            return '/'.join(map(str, [self.mn, self.avg, self.mx]))
  
        fdv = None
        flr = None
        ftd = None
        ir  = None
        loss = None
        status = None
  
        def __init__(self, fdv, flr, ftd, ir, loss, status):
          self.fdv    = fdv
          self.flr    = flr
          self.ftd    = ftd
          self.ir     = ir
          self.loss   = loss
          self.status = status
  
        @classmethod
        def from_dict(cls, dct):
          return cls(
            cls.NumberRange.from_dict(dct['fdv_ms']),
            dct['flr'],
            cls.NumberRange.from_dict(dct['ftd_ms']),
            cls.NumberRange.from_dict(dct['ir']),
            dct['loss'],
            dct['status'])
  
        def to_table_row(self, strings):
          return [
            strings['statuses'][self.status.lower()],
            self.ir,
            str(self.flr),
            str(self.ftd) + ' ' + strings['units']['ms'],
            str(self.fdv) + ' ' + strings['units']['ms']
          ]

      def __init__(self, configuration, performance):
        self.configuration = configuration
        self.performance   = performance

      @classmethod
      def from_dict(cls, dct):
        return cls(
          cls.Configuration.from_dict(dct['configuration'])
            if not dct['configuration'] is None else None,
          cls.Performance.from_dict(dct['performance'])
            if not dct['performance'] is None else None
        )

    class Settings:

      header = None
      cir = None
      eir = None
      frame = None
      sac = dict (
        fdv = None,
        flr = None,
        ftd = None
      )

      def __init__(self, header, cir, eir, frame, sac):
        self.header = header
        self.cir = cir
        self.eir = eir
        self.frame = frame
        self.sac['fdv'] = sac['fdv_ms']
        self.sac['flr'] = sac['flr']
        self.sac['ftd'] = sac['ftd_ms']

      @classmethod
      def from_dict(cls, dct):
        return cls(
          HeaderInfo.from_dict(dct['header']),
          RateInfo.from_dict(dct['config']['cir']),
          RateInfo.from_dict(dct['config']['eir']),
          dct['config']['frame'],
          dct['config']['sac']
        )

      def sac_to_report(self, strings, level = 4):
        title = Paragraph(strings['y1564']['sac_settings'], level)
        dct = Dictionary([
          (strings['y1564']['fdv'], str(self.sac['fdv']) + ' ' +
             strings['units']['ms']),
          (strings['y1564']['ftd'], str(self.sac['ftd']) + ' ' +
             strings['units']['ms']),
          (strings['y1564']['loss'], str(self.sac['flr']))
        ])
        return Group([title, dct])

      def to_report(self, strings, level = 3):
        dct = Dictionary([
          (strings['common']['frame'], str(self.frame) +
             strings['units']['B']),
          (strings['y1564']['cir'], self.cir.to_string(strings)),
          (strings['y1564']['eir'], self.eir.to_string(strings))
        ])
        hdr = self.header.to_report(strings, level + 1)
        sac = self.sac_to_report(strings, level + 1)
        return Group([dct, hdr, sac])

      def to_table_row(self, strings):
        return [str(self.frame) + strings['units']['B'],
          self.cir.to_string(strings)]

    def __init__(self, settings, results):
      self.settings = settings
      self.results  = results

    @classmethod
    def from_dict(cls, dct):
      return cls(
        cls.Settings.from_dict(dct['settings']),
        cls.Results.from_dict( dct['results'])
      )

  class Settings:

    topology = None
    configuration = dict (
      step_duration = None,
      cir_step_count = None,
      eir_enabled = None,
      tp_enabled = None
    )
    performance = dict (
      enabled = None,
      duration = None
    )
    streams = None

    def __init__(self, topology, configuration, performance, streams):
      self.topology = topology

      if not configuration is None:
        self.configuration = dict (
          step_duration  = configuration['step_duration' ],
          cir_step_count = configuration['cir_step_count'],
          eir_enabled    = configuration['eir_enabled'   ],
          tp_enabled     = configuration['tp_enabled'    ]
        )
      else:
        self.configuration = None

      if not performance is None:
        self.performance = dict (
          enabled  = performance['enabled'],
          duration = performance['duration']
        )
      else:
        self.performance = None

      self.streams = self.streams

    def configuration_enabled(self):
      return (
        (not self.configuration is None) and (
          self.configuration['cir_step_count'] != 0 or
          self.configuration['eir_enabled'] or
          self.configuration[ 'tp_enabled']
        )
      )

    def performance_enabled(self):
      return ((not self.performance is None) and
        self.performance['enabled'])

    @classmethod
    def from_dict(cl, dct):
      return cl(
        None if dct['topology'] is None else
          TopologyInfo.from_dict(dct['topology']),
        None if dct['config'] is None else
          dct['config']['configuration'],
        None if dct['config'] is None else
          dct['config']['performance'],
        None if dct['config'] is None else
          dct['config']['streams']
      )

    def to_report(self, strings, level = 4):
      grp = []

      tit = Paragraph(strings['common']['settings'], level)
      grp.append(tit)

      if not self.topology is None:
        grp.append(elf.topology.to_report(strings, level + 1))

      if self.configuration_enabled():
        conf_str = strings['common']['enabled'] + ' ('

        stages = []
        if self.configuration['cir_step_count'] != 0:
          stages.append(strings['y1564']['cir'] + ' 1-' +
            str(self.configuration['cir_step_count']))
        if self.configuration['eir_enabled']:
          stages.append(strings['y1564']['eir'])
        conf_str += ', '.join(stages) + ')'

        conf_str += ', ' + strings['y1564'][
          'w_tp' if self.configuration['tp_enabled'] else 'wo_tp'
        ]

        conf_str += ', ' + strings['y1564']['with_step'] + ' ' + (
          str(self.configuration['step_duration']) + ' ' +
          strings['units']['s']
        )

      else:
        conf_str = strings['common']['disabled'] 

      if self.performance_enabled():
        perf_str = ' '.join([
          strings['common']['enabled'] + ',',
          strings['y1564']['with_duration'],
          str(self.performance['duration']),
          strings['units']['s']
        ])
      else:
        perf_str = strings['common']['disabled']

      if not self.streams is None:
        dct = Dictionary([
          (strings['y1564']['configuration'], conf_str),
          (strings['y1564']['performance'],   perf_str),
          (strings['y1564']['stream_count'],  self.streams)
        ])
        grp.append(dct)

      return Group(grp)

  def __init__(self, info, ports, settings, streams):
    self.info     = info
    self.ports    = ports
    self.settings = settings
    self.streams  = streams

  @classmethod
  def results_hook(cls, dct):
    return map(lambda x : cls.Streams.from_dict(x), dct['streams'])

  def to_report(self, strings, level = 1):

    grp = []

    title = Paragraph(strings['y1564']['name'], level)
    grp.append(title)

    grp.extend(map(lambda x : x.to_report(strings, level + 1),
      filter(None, [self.info, self.ports, self.settings])))

    for i in range(len(self.streams)):
      sttit = Paragraph(strings['y1564']['service'] + ' ' + str(i+1),
        level + 2)
      stset = self.streams[i].settings.to_report(strings,
        level + 2)
      grp.extend([sttit, stset])

    grp.append(Paragraph(strings['common']['results'], level + 1))

    if self.settings.configuration_enabled():
      grpin = []

      for i in range(len(self.streams)):
        stream_tit = Paragraph(strings['y1564']['configuration'] +
          ', ' + strings['y1564']['service'] + ' ' + str(i+1),
          level + 2)

        stream_res = self.streams[i].results.configuration.to_report(
          strings, level + 2)

        grpin.extend([stream_tit, stream_res])

      grp.extend(grpin)

    if self.settings.performance_enabled():
      title = Paragraph(strings['y1564']['performance'], level + 2)

      rows = []
      for i in range(len(self.streams)):
        row = self.streams[i].results.performance.to_table_row(
          strings)
        frame = self.streams[i].settings.frame
        cir = (self.streams[i].settings.cir.to_string(strings))
        rows.append([frame, cir] + row)

      hdr = [
       strings['common']['frame'],
       strings['y1564']['cir'],
       strings['common']['status'],
       strings['y1564']['ir'],
       strings['y1564']['loss'],
       strings['y1564']['ftd'],
       strings['y1564']['fdv']
      ]
      perf_res = Table([hdr] + rows)

      grp.extend([title, perf_res])

    return Group(grp)

class VcTest (MtkTest):

  class Results:

    a = None
    b = None

    class Measurement:

      attenuation = None
      crossover   = None
      distance    = None
      pair        = None
      polarity    = None
      status      = None

      def __init__(self, attenuation, crossover, distance, pair,
          polarity, status):
        self.attenuation = attenuation
        self.crossover   = crossover
        self.distance    = distance
        self.pair        = pair
        self.polarity    = polarity
        self.status      = status

      @classmethod
      def from_dict(cls, dct):
        return cls(
          dct['attenuation'],
          dct['crossover'],
          dct['distance'],
          dct['pair'],
          dct['polarity'],
          dct['status']
	)

      def to_table_row(self, strings):
        cell = strings['vct']['statuses'][self.status]
        if self.status == 'norm':
          cell = ', '.join([
            cell,
            strings['vct']['polarity'][self.polarity] + ' ' +
              strings['vct']['polarity']['name'],
            strings['vct']['crossover']['mdix']
              if self.crossover == 'MDI-X' else
              strings['vct']['crossover']['mdi']
          ])
        elif self.status == 'short' or self.status == 'open':
          cell = ', '.join([
            cell,
            strings['vct']['attenuation'] + ' ' +
              str(self.attenuation) + ' ' + strings['units']['db'],
            strings['vct']['distance'] + ' ' +
              str(self.distance) + ' ' + strings['units']['m']
          ])
        elif self.status != 'fail' and self.status != 'test':
          raise KeyError(self.status + ' is not a VCT status')
        return [self.pair, cell]

    def __init__(self, a, b):
      self.a = a
      self.b = b

    @classmethod
    def from_dict(cls, dct):
      return cls(
        map(lambda x : cls.Measurement.from_dict(x), dct['a']),
        map(lambda x : cls.Measurement.from_dict(x), dct['b'])
      )

    def to_report(self, strings, level = 4):
      grp = []

      title = Paragraph(strings['common']['results'], level)
      grp.append(title)

      for p in ['a', 'b']:
        ptit = Paragraph(strings['common']['port'][p], level + 1)
        pres = map(lambda x : x.to_table_row(strings),
          getattr(self, p))
        ptbl = Table(pres)

        grp.extend([ptit, ptbl])

      return Group(grp)

  @classmethod
  def settings_hook(cls, dct):
    return None

  def to_report(self, strings, level = 1):
    return self._to_report(strings['vct']['name'],
      strings, level)

class MtkTests:

  bert = None
  rfc2544 = None
  y1564 = None
  vct = None

  def __init__(self, bert, rfc2544, y1564, vct):
    self.bert = bert
    self.rfc2544 = rfc2544
    self.y1564 = y1564
    self.vct = vct

  @staticmethod
  def from_dict(dct):
    return MtkTests(
      map(lambda x : BerTest.from_dict(x), dct['bert']),
      map(lambda x : Rfc2544Test.from_dict(x), dct['rfc2544']),
      None if dct['y1564'] is None
        else Y1564Test.from_dict(dct['y1564']),
      None if dct['vct'] is None else VcTest.from_dict(dct['vct'])
    )

  def to_report(self, strings, level = 2):
    return Group(map(lambda x : x.to_report(strings, level + 1),
      filter(None, self.bert + self.rfc2544 + [self.y1564] +
        [self.vct])))

class Device:

  name   = None
  sn     = None
  firmware = None

  def __init__(self, name, sn, firmware):
    self.name = name
    self.sn = sn
    self.firmware = firmware

  @classmethod
  def from_dict(cls, dct):
    return cls(dct['name'], dct['sn'], dct['firmware'])

  def to_report(self, strings, level = 1):
    title = Paragraph(strings['device']['device_info'], level)
    dct = Dictionary([
      (strings['device']['name'],
         strings['device']['names'][self.name]),
      (strings['device']['manufacturer'],
         strings['device']['manufacturers']['default']),
      (strings['device']['sn'], self.sn),
      (strings['device']['firmware'], self.firmware)
    ])
    return Group([title, dct])

class MtkReport:

  device = None
  tests  = None

  def __init__(self, device, tests):
    self.device = device
    self.tests = tests

  @staticmethod
  def from_dict(dct):
    return MtkReport(Device.from_dict(dct['device']),
      MtkTests.from_dict(dct['tests']))

  def to_report(self, strings, level = 0):
    grp = []
    title = Paragraph(strings['general']['title'], 1)
    grp.append(title)

    text = Paragraph(strings['general']['report_saved_on'] +
      ' ' + datetime.now().strftime('%d %b %Y %H:%M:%S'))
    grp.append(text)

    device_info = self.device.to_report(strings, level + 1)
    grp.append(device_info)

    tests = self.tests.to_report(strings, level + 1)
    grp.append(tests)

    return Group(grp)

