from base64 import b64encode
from reportlab.lib import colors
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.pagesizes import letter
import reportlab.graphics.shapes as rlgraphsh
import reportlab.graphics.charts.barcharts as rlgraphch
import docx, docx.shared
import reportlab.platypus
import tempfile, shutil
from subprocess import call, STDOUT
from os import devnull

from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont

enc = 'UTF-8'

pdfmetrics.registerFont(TTFont('DejaVuSans', 'DejaVuSans.ttf',enc))
pdfmetrics.registerFont(TTFont('DejaVuSans-Bold', 'DejaVuSans-Bold.ttf',enc))

class FormatTypes:
  HTML = 1
  PDF  = 2
  DOCX = 3

class DocStructure:

  def __init__(self, data):
    return NotImplemented

  def format_full(self, fmttype, outfile = None):
    return NotImplemented

  def format_inline(self, fmttype, ctxt):
    if fmttype == FormatTypes.HTML:
      return self._format_html(ctxt)
    elif fmttype == FormatTypes.PDF:
      return self._format_pdf(ctxt)
    elif fmttype == FormatTypes.DOCX:
      return self._format_docx(ctxt)

  def update_ctxt(self, old_ctxt):
    old_ctxt['stack'].append(type(self))
    return old_ctxt

class Document(DocStructure):

  data = ''

  def __init__(self, data):
    if not isinstance(data, DocStructure):
      raise TypeError('data is not an instance of DocStructure')
    self.data = data

  @staticmethod
  def flatten(lst):
    return sum(map(lambda x : Document.flatten(x) if type(x) is list
               else [x], lst), [])

  def format_full(self, fmttype, outfile = None):
    ctxt = dict(
      stack = [],
      cleanup = []
    )
    t = Document.flatten(self.format_inline(fmttype, ctxt))
    if fmttype == FormatTypes.HTML:
      if outfile != None:
        with open(outfile, 'w') as f:
          f.write(''.join(t).encode('UTF-8'))
      return t
    elif fmttype == FormatTypes.PDF:
      doc = reportlab.platypus.SimpleDocTemplate(outfile,
                                           pagesize=letter)
      doc.build(t)
    elif fmttype == FormatTypes.DOCX:
      doc = docx.Document()
      for fn in t:
        fn(doc)
      doc.save(outfile)
    for dr in ctxt['cleanup']:
      shutil.rmtree(dr)

  def _format_html(self, ctxt):
    st  = '<style>table tr td {border: 1px solid black;}</style>'
    st += '<meta charset="UTF-8" />'
    return ['<!DOCTYPE html>', '<html><head>' + st + '</head><body>',
           self.data.format_inline(FormatTypes.HTML,
           self.update_ctxt(ctxt)), '</body></html>']

  def _format_pdf(self, ctxt):
    return self.data.format_inline(FormatTypes.PDF,
           self.update_ctxt(ctxt))

  def _format_docx(self, ctxt):
    return self.data.format_inline(FormatTypes.DOCX,
           self.update_ctxt(ctxt))

class Paragraph(DocStructure):

  text = ''
  level = 0

  def __init__(self, text, level = 0, options = None):
    self.text = text
    self.level = level

  def get_level(self, ctxt):
    return self.level

  def _format_html(self, ctxt):
    lvl = self.get_level(ctxt)
    if lvl == 0:
      return ['<p>', self.text, '</p>']
    else:
      lvl = str(lvl)
      return ['<h', lvl, '>', self.text, '</h', lvl, '>']

  def _format_pdf(self, ctxt):
    text = """<font name=DejaVuSans>%s</font>""" % self.text
    styles = getSampleStyleSheet()
    lvl = self.get_level(ctxt)
    if lvl == 0:
      style = styles['Normal']
    else:
      style = styles['Heading' + str(lvl)]
    return reportlab.platypus.Paragraph(text, style)

  def _format_docx(self, ctxt):
    return (lambda doc : doc.add_paragraph(self.text)
            if self.get_level(ctxt) == 0
            else doc.add_heading(self.text,
                                 level = self.get_level(ctxt)))

class TextUnit(DocStructure):

  text = ''

  def __init__(self, text, style = None):
    self.text = text

  def format_inline(self, fmttype, ctxt):
    return (self.text.format_inline(fmttype, self.update_ctxt(ctxt))
      if isinstance(self.text, DocStructure) else unicode(self.text))

class Group(DocStructure):

  data = []

  def __init__(self, data):
    self.data = data

  def format_inline(self, fmttype, ctxt):
    return sum(map(lambda x :
           [x.format_inline(fmttype, self.update_ctxt(ctxt))],
            self.data), [])

class Dictionary(DocStructure):

  data = ''

  def __init__(self, data):
    self.data = map(lambda r : (TextUnit(r[0]), TextUnit(r[1])), data)

  def _data_format(self, fmttype, ctxt):
    nctxt = self.update_ctxt(ctxt)
    return map(
      lambda d : (d[0].format_inline(fmttype, nctxt),
                  d[1].format_inline(fmttype, nctxt)), self.data)

  def _format_html(self, ctxt):
    return ['<dl>', sum(
           map(lambda x :
           ['<dt>', x[0], '</dt>', '<dd>', x[1], '</dd>'],
           self._data_format(FormatTypes.HTML, ctxt)), []), '</dl>']

  def _format_pdf(self, ctxt):
    f = map(lambda x : [x[0], x[1]],
      self._data_format(FormatTypes.PDF, ctxt))
    t = reportlab.platypus.Table(f, hAlign = 'LEFT')
    t.setStyle(reportlab.platypus.TableStyle([
      ('FONTNAME', (0,0), (-1,-1), 'DejaVuSans')
    ]))
    return t

  def _format_docx(self, ctxt):
    if len(self.data) == 0:
      return []

    def fn(doc):
      tbl = doc.add_table(rows = 0, cols = 2)
      for r in self._data_format(FormatTypes.DOCX, ctxt):
        cls = tbl.add_row().cells
        cls[0].text = r[0]
        cls[1].text = r[1]

    return fn

class Image(DocStructure):

  path = ''
  width = None
  height = None

  def __init__(self, path, width = None, height = None):
    self.path = path
    self.width = width
    self.height = height

  @staticmethod
  def _format_html_raw(path, width = None, height = None):
    with open(path, 'r') as img:
      data = b64encode(img.read())
    wtext = htext = ""
    src_text = 'src="data:image/png;base64,' + data + '"'
    if (width != None):
      wtext = 'width="' + str(width) + '"'
    if (height != None):
      htext = 'height="' + str(height) + '"'
    img = ' '.join(['<img', src_text, wtext, htext, '/>'])
    return img

  def _format_html(self, ctxt):
    return Image._format_html_raw(self.path, self.width, self.height)

  @staticmethod
  def _format_pdf_raw(path, width = None, height = None):
    img = reportlab.platypus.Image(path)
    if (height != None) and (width != None):
      img._restrictSize(height, width)
    return img

  def _format_pdf(self, ctxt):
    return Image._format_pdf_raw(self.path, self.width, self.height)

  @staticmethod
  def _format_docx_raw(path, width = None, height = None):
    return (lambda doc : doc.add_picture(path,
      None if width  is None else docx.shared.Pt(width),
      None if height is None else docx.shared.Pt(height)))

  def _format_docx(self, ctxt):
    return Image._format_docx_raw(self.path, self.width, self.height)

class Table(DocStructure):

  data = []

  def __init__(self, data):
    self.data = []
    for r in data:
      rn = map(lambda x : x if isinstance(x, DocStructure)
        else TextUnit(x), r)
      self.data.append(rn)

  def _format_html(self, ctxt):
    t = []
    for r in self.data:
      formatted = map(lambda x : x.format_inline(FormatTypes.HTML,
                                 self.update_ctxt(ctxt)), r)
      cells = map(lambda x : ['<td>', x, '</td>'], formatted)
      t += ['<tr>', cells, '</tr>']
    return ['<table>', t, '</table>']

  def _format_pdf(self, ctxt):
    k = []
    for r in self.data:
      rn = map(lambda x : x.format_inline(FormatTypes.PDF,
                                          self.update_ctxt(ctxt)), r)
      k.append(rn)
    t = reportlab.platypus.Table(k)
    t.setStyle(reportlab.platypus.TableStyle([
      ('FONTNAME', (0,0), (-1,-1), 'DejaVuSans'),
      ('INNERGRID', (0,0), (-1,-1), 0.25, colors.black),
      ('BOX', (0,0), (-1,-1), 0.25, colors.black)]))
    return t

  def _format_docx(self, ctxt):
    if len(self.data) == 0:
      return []

    def fn(doc):
      tbl = doc.add_table(rows = 0, cols = len(self.data[0]))
      for r in self.data:
        cls = tbl.add_row().cells
        for i in range(0, len(r)):
          cls[i].text = r[i].format_inline(FormatTypes.DOCX,
                        self.update_ctxt(ctxt))

    return fn

class Graphic(DocStructure):

  data   = None
  script = None
  width  = None
  height = None
  labels = None

  def __init__(self, data, labels, script, width = 600, height = 450):
    self.data   = data
    self.script = script
    self.width  = width
    self.height = height
    self.labels = labels

  @staticmethod
  def _format_image_raw(data, labels, script, dr, fmt, width, height):
    fname = dr + '/' + 'data'
    with open(fname, 'w+') as f:
      f.write(data.encode('utf-8'))
    outname = dr + '/' + 'output'
    with open(devnull, 'w') as FNULL:
#      print ' '.join(["gnuplot", "-e", "datasrc='" + fname + "'",
#                       "-e", "filename='" + outname + "'",
#                       "-e", "driver='" + fmt + "'",
#                       "-e", "xlbl='" + labels[0] + "'",
#                       "-e", "ylbl='" + labels[1] + "'",
#           script])
      call(["gnuplot", "-e", "datasrc='" + fname + "'",
                       "-e", "filename='" + outname + "'",
                       "-e", "driver='" + fmt + "'",
                       "-e", "xlbl='" + labels[0] + "'",
                       "-e", "ylbl='" + labels[1] + "'",
           script], stdout=FNULL, stderr=STDOUT)
    return outname

  def _format_image(self, dr, fmt, width = None, height = None):
    w = self.width  if width  is None else width
    h = self.height if height is None else height
    return Graphic._format_image_raw(self.data, self.labels,
      self.script, dr, fmt, w, h)

  def _format_html(self, ctxt):
    dr = tempfile.mkdtemp()
    outfile = self._format_image(dr, 'png')
    img = Image._format_html_raw(outfile, self.width, self.height)
    shutil.rmtree(dr)
    return img

  def _format_pdf(self, ctxt):
    dr = tempfile.mkdtemp()
    outfile = self._format_image(dr, 'png')
    img = Image._format_pdf_raw(outfile, self.width, self.height)
    ctxt['cleanup'].append(dr)
    return img

  def _format_docx(self, ctxt):
    
    def fn(doc):
      dr = tempfile.mkdtemp()
      outfile = self._format_image(dr, 'png')
      oldFn = Image._format_docx_raw(outfile,
        None if self.width  is None else 0.6 * self.width,
        None if self.height is None else 0.6 * self.height)
      oldFn(doc)
      shutil.rmtree(dr)

    return fn
